/**
* Copyright (C) Mellanox Technologies Ltd. 2001-2017.  ALL RIGHTS RESERVED.
*
* See file LICENSE for terms.
*/

#ifndef UCS_TIME_DEF_H
#define UCS_TIME_DEF_H


/**
 * @ingroup UCS_RESOURCE
 *
 * UCS time units.
 * These are not necessarily aligned with metric time units.
 * MUST compare short time values with UCS_SHORT_TIME_CMP to handle wrap-around.
 */
typedef unsigned long   ucs_time_t;


#endif /* UCS_TIME_DEF_H */
