{"UCX_CONFIG_H", UCS_PP_MAKE_STRING()},
{"ENABLE_ASSERT", UCS_PP_MAKE_STRING(1)},
{"ENABLE_DEBUG_DATA", UCS_PP_MAKE_STRING(0)},
{"ENABLE_MT", UCS_PP_MAKE_STRING(0)},
{"ENABLE_PARAMS_CHECK", UCS_PP_MAKE_STRING(1)},
{"ENABLE_SYMBOL_OVERRIDE", UCS_PP_MAKE_STRING(1)},
{"HAVE_ATTRIBUTE_NOOPTIMIZE", UCS_PP_MAKE_STRING(1)},
{"HAVE_CMA", UCS_PP_MAKE_STRING(1)},
{"HAVE_CUDA", UCS_PP_MAKE_STRING(0)},
{"HAVE_DECL_ASPRINTF", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_BASENAME", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_CPU_ISSET", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_CPU_ZERO", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_ETHTOOL_CMD_SPEED", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_FMEMOPEN", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_F_SETOWN_EX", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_IBV_WC_STATUS_STR", UCS_PP_MAKE_STRING(0)},
{"HAVE_DECL_SPEED_UNKNOWN", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_STRDUPA", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL_STRERROR_R", UCS_PP_MAKE_STRING(1)},
{"HAVE_DECL___PPC_GET_TIMEBASE_FREQ", UCS_PP_MAKE_STRING(0)},
{"HAVE_DLFCN_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_GDR_COPY", UCS_PP_MAKE_STRING(0)},
{"HAVE_HW_TIMER", UCS_PP_MAKE_STRING(1)},
{"HAVE_INTTYPES_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_KNEM", UCS_PP_MAKE_STRING(0)},
{"HAVE_LIBRT", UCS_PP_MAKE_STRING(1)},
{"HAVE_MALLOC_HOOK", UCS_PP_MAKE_STRING(1)},
{"HAVE_MEMORY_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_ROCM", UCS_PP_MAKE_STRING(0)},
{"HAVE_STDINT_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_STDLIB_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_STRERROR_R", UCS_PP_MAKE_STRING(1)},
{"HAVE_STRINGS_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_STRING_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_STRUCT_DL_PHDR_INFO", UCS_PP_MAKE_STRING(1)},
{"HAVE_SYS_STAT_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_SYS_TYPES_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_SYS_UIO_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_UCM_PTMALLOC286", UCS_PP_MAKE_STRING(1)},
{"HAVE_UNISTD_H", UCS_PP_MAKE_STRING(1)},
{"HAVE_XPMEM", UCS_PP_MAKE_STRING(0)},
{"LT_OBJDIR", UCS_PP_MAKE_STRING(".libs/")},
{"NVALGRIND", UCS_PP_MAKE_STRING(1)},
{"PACKAGE", UCS_PP_MAKE_STRING("ucx")},
{"PACKAGE_BUGREPORT", UCS_PP_MAKE_STRING("")},
{"PACKAGE_NAME", UCS_PP_MAKE_STRING("ucx")},
{"PACKAGE_STRING", UCS_PP_MAKE_STRING("ucx 1.4")},
{"PACKAGE_TARNAME", UCS_PP_MAKE_STRING("ucx")},
{"PACKAGE_URL", UCS_PP_MAKE_STRING("")},
{"PACKAGE_VERSION", UCS_PP_MAKE_STRING("1.4")},
{"STDC_HEADERS", UCS_PP_MAKE_STRING(1)},
{"STRERROR_R_CHAR_P", UCS_PP_MAKE_STRING(1)},
{"UCS_MAX_LOG_LEVEL", UCS_PP_MAKE_STRING(UCS_LOG_LEVEL_TRACE_POLL)},
{"UCX_CONFIGURE_FLAGS", UCS_PP_MAKE_STRING("--disable-numa")},
{"VERSION", UCS_PP_MAKE_STRING("1.4")},
{"restrict", UCS_PP_MAKE_STRING(__restrict)},
